"""
Module that summarizes an assignment group by disjunctive maximum.

Disjunctive maximum means: 

  1) At least one assignment must have a non-F grade.
  2) If there are more than one assignment with a non-F grade, we take the 
     maximum as the grade. A--E are valued higher than P. The grade F is valued 
     the lowest.

We fail if there is an assignment which doesn't have A--F or P/F grading 
scales.
"""

import datetime as dt
from canvaslms.cli import results

def summarize(user, assignments_list):
  """Extracts user's submissions for assignments in assingments_list to 
  summarize results into one grade and a grade date. Summarize by disjunctive 
  maximum."""

  grades = []
  dates = []
  graders = []

  for assignment in assignments_list:
    submission = assignment.get_submission(user,
                                           include=["submission_history"])
    submission.assignment = assignment
    graders += results.all_graders(submission)

    grade = submission.grade

    if grade is None:
      grade = "F"

    grades.append(grade)

    grade_date = submission.submitted_at or submission.graded_at

    if grade_date:
      grade_date = dt.date.fromisoformat(grade_date.split("T")[0])
      dates.append(grade_date)

  return (grade_max(grades), max(dates), graders)
def grade_max(grades):
  """Takes a list of A--E/P--F grades, returns the maximum."""
  num_grades = map(grade_to_int, grades)
  max_grade = max(num_grades)
  return int_to_grade(max_grade)

def grade_to_int(grade):
  grade_map = {"F": -1, "P": 0, "E": 1, "D": 2, "C": 3, "B": 4, "A": 5}
  return grade_map[grade]

def int_to_grade(int_grade):
  grade_map_inv = {-1: "F", 0: "P", 1: "E", 2: "D", 3: "C", 4: "B", 5: "A"}
  return grade_map_inv[int_grade]

def summarize_group(assignments_list, users_list):
  """Summarizes a particular set of assignments (assignments_list) for all
  users in users_list"""

  for user in users_list:
    grade, grade_date, graders = summarize(user, assignments_list)
    yield [user, grade, grade_date, *graders]



"""A module that modifies the classes of the canvasapi package"""

import importlib
import inspect
import sys

def make_classes_comparable():
  """Improves the classes by adding __eq__ and __hash__ methods"""
  def canvas_comparable(cls):
    def is_equal(self, other):
      """Tests if Canvas objects self and other refer to the same object"""
      return type(self) == type(other) and self.id == other.id

    cls.__eq__ = is_equal
    return cls
  def canvas_hashable(cls):
    def canvas_hash(self):
      """Returns a hash suitable for Canvas objects"""
      return hash(type(self)) ^ hash(self.id)

    cls.__hash__ = canvas_hash
    return cls
  # classes to improve in each module
  CANVASAPI_CLASSES = {
    "assignment": ["Assignment", "AssignmentGroup"],
    "submission": ["Submission"],
    "user": ["User"]
  }
  canvasapi_modules = {}

  # import all modules
  for module_name in CANVASAPI_CLASSES:
    canvasapi_modules[module_name] = \
      importlib.import_module(f"canvasapi.{module_name}")
  for module_name, module in canvasapi_modules.items():
    module_members = inspect.getmembers(module)
    for obj_name, obj in module_members:
      if obj_name in CANVASAPI_CLASSES[module_name]:
        canvas_comparable(obj)
        canvas_hashable(obj)
def make_useful_user_dunder_str():
  """Improves the user class by changing __str__"""
  def name_and_login(self):
    return f"{self.name} <{self.login_id}>"
  import canvasapi.user
  canvasapi.user.User.__str__ = name_and_login

# Loads all hacks
this_module = sys.modules[__name__]

# automatically execute all functions in this module
for _, function in inspect.getmembers(this_module, inspect.isfunction):
  function()



"""
This module is the same as `canvaslms.grades.conjunctavg` except that any 
submissions with grades other than A--F and P/F are treated as P. For instance, 
numeric grades (like points). Also, all submissions must have a date. This 
makes this module useful for including mandatory, ungraded surveys.
"""

import datetime as dt
from canvaslms.grades.conjunctavg import a2e_average

def summarize(user, assignments_list):
  """
  Extracts user's submissions for assignments in assingments_list to summarize 
  results into one grade and a grade date. Summarize by conjunctive average.

  If some submission lacks date, return ("F", None).
  """

  pf_grades = []
  a2e_grades = []
  recent_date = dt.date(year=1970, month=1, day=1)

  for assignment in assignments_list:
    submission = assignment.get_submission(user)
    grade = submission.grade

    if grade is None:
      grade = "F"

    if grade in "ABCDE":
      a2e_grades.append(grade)
    elif grade in "PF":
      pf_grades.append(grade)
    else:
      pf_grades.append("E")

    grade_date = submission.submitted_at or submission.graded_at

    if not grade_date:
      return ("F", None)
    else:
      grade_date = dt.date.fromisoformat(grade_date.split("T")[0])

    if grade_date > recent_date:
      recent_date = grade_date

  if not all(map(lambda x: x == "P", pf_grades)):
    return ("F", recent_date)

  if a2e_grades:
    return (a2e_average(a2e_grades), recent_date)
  return ("P", recent_date)

def summarize_group(assignments_list, users_list):
  """Summarizes a particular set of assignments (assignments_list) for all
  users in users_list"""

  for user in users_list:
    grade, grade_date = summarize(user, assignments_list)
    yield (user, grade, grade_date)



"""
Package containing modules to summarize assignment groups in different ways.

For a module to be used with the `canvaslms results -S module` option, the 
module must fulfil the following:

  1) It must contain a function named `summarize_group`.
  2) `summarize_group` must take two arguments:

      I) `assignment_list`, a list of `canvasapi.assignment.Assignment` 
      objects.

      II) `users_list`, a list of `canvasapi.user.User` objects.

  3) The return value should be a list of tuples. Each tuple should have the 
     form `(user, grade, grade date)`.

See the built-in modules below.
"""



import appdirs
import canvaslms.cli
import json
import keyring
import os

dirs = appdirs.AppDirs("canvaslms", "dbosk@kth.se")

def login_command(config, canvas, args):
  """Guides the user to update credentials"""

  print("Enter the hostname for Canvas, "
    "e.g. 'canvas.kth.se' or 'kth.instructure.com'.")
  hostname = input("Canvas hostname: ")

  print(f"""
Open

  https://{hostname}/profile/settings

in your browser. Scroll down to approved integrations and click the
'+ New access token' button. Fill in the required data and click the
'Generate token' button. Enter the token here.
""")

  token = input("Canvas token: ")

  try:
    keyring.set_password("canvaslms", "hostname", hostname)
    keyring.set_password("canvaslms", "token", token)
  except:
    canvaslms.cli.warn(f"You don't have a working keyring. "
                       f"Will write hostname and token to config file "
                       f"{args.config_file}.")

    config["canvas"]["host"] = hostname
    config["canvas"]["access_token"] = token

    canvaslms.cli.update_config_file(config, args.config_file)

def load_credentials(config):
  """Load credentials from keyring, environment or config dictionary"""
  try:
    hostname = keyring.get_password("canvaslms", "hostname")
    token = keyring.get_password("canvaslms", "token")
    if hostname and token:
      return hostname, token
  except:
    pass

  try:
    hostname = os.environ["CANVAS_SERVER"]
    token = os.environ["CANVAS_TOKEN"]
    return hostname, token
  except KeyError:
    pass

  try:
    return config["canvas"]["host"], config["canvas"]["access_token"]
  except KeyError:
    pass

  return None, None

def add_command(subp):
  """Adds the login command to argparse parser"""
  login_parser = subp.add_parser("login",
    help="Manage login credentials",
    description=f"""
Manages the user's Canvas login credentials. There are three ways to supply the 
login credentials, in order of priority:

1) Through the system keyring: Just run `canvaslms login` and you'll be guided 
   to enter the credentials (server name and token) and they will be stored in 
   the keyring.

2) Through the environment: Just set the environment variables CANVAS_SERVER 
   and CANVAS_TOKEN.

3) Through the configuration file: Just write

      {{
        "canvas": {{
          "host": "the actual hostname",
          "access_token": "the actual token"
        }}
      }}

   to the file {dirs.user_config_dir}/config.json (default, or use the -f 
   option, see `canvaslms -h`).
""")
  login_parser.set_defaults(func=login_command)



from canvaslms.cli import submissions, users
import webbrowser

def grade_command(config, canvas, args):
  submission_list = submissions.process_submission_options(canvas, args)
  results = {}
  if args.grade:
    results["submission"] = {"posted_grade": args.grade}
  if args.message:
    results["comment"] = {"text_comment": args.message}
  if not args.grade and not args.message:
    for submission in submission_list:
      webbrowser.open(submissions.speedgrader(submission))
  else:
    for submission in submission_list:
      submission.edit(**results)

def add_command(subp):
  """Adds grade command to the argparse subparser subp"""
  grade_parser = subp.add_parser("grade",
    help="Grades assignments (hic sunt dracones!)",
    description="Grades assignments. ***Hic sunt dracones [here be dragons]***: "
      "the regex matching is very powerful, "
      "be certain that you match what you think!")
  grade_parser.set_defaults(func=grade_command)
  submissions.add_submission_options(grade_parser, required=True)
  grade_options = grade_parser.add_argument_group(
    "arguments to set the grade and/or comment, "
    "if none given, opens SpeedGrader")
  grade_options.add_argument("-g", "--grade",
    help="The grade to set for the submissions")
  grade_options.add_argument("-m", "--message",
    help="A comment to the student")



import canvaslms.cli
import canvaslms.cli.assignments as assignments
import canvaslms.cli.courses as courses
import canvaslms.cli.submissions as submissions
import canvaslms.hacks.canvasapi

import argparse
import csv
import canvasapi.submission
import datetime as dt
import importlib
import importlib.machinery
import importlib.util
import os
import pathlib
import re
import sys

def results_command(config, canvas, args):
  output = csv.writer(sys.stdout, delimiter=args.delimiter)

  assignments_list = assignments.process_assignment_option(canvas, args)

  if args.assignment_group != "":
    results = summarize_assignment_groups(canvas, args)
  else:
    results = summarize_assignments(canvas, args)

  for result in results:
    if not args.include_Fs and result[3][0] == "F":
      continue
    output.writerow(result)
def summarize_assignments(canvas, args):
  """
  Turn submissions into results:
  - canvas is a Canvas object,
  - args is the command-line arguments, as parsed by argparse.
  """

  courses_list = courses.process_course_option(canvas, args)

  users_list = []
  for course in courses_list:
    for user in course.get_users(enrollment_type=["student"]):
      users_list.append(user)

  assignments_list = assignments.process_assignment_option(canvas, args)
  submissions_list = submissions.filter_submissions(
    submissions.list_submissions(assignments_list,
                                 include=["submission_history"]),
    users_list)

  for submission in submissions_list:
    if submission.grade is not None:
      yield [
        submission.assignment.course.course_code,
        submission.assignment.name,
        submission.user.integration_id,
        submission.grade,
        round_to_day(submission.submitted_at or submission.graded_at),
        *all_graders(submission)
      ]
def round_to_day(timestamp):
  """
  Takes a Canvas timestamp and returns the corresponding datetime.date object.
  """
  return dt.date.fromisoformat(timestamp.split("T")[0])
def all_graders(submission):
  """
  Returns a list of everyone who participated in the grading of the submission. 
  I.e. also those who graded previous submissions, when submission history is 
  available.
  """
  graders = []

  for prev_submission in submission.submission_history:
    prev_submission = canvasapi.submission.Submission(
      submission._requester, prev_submission)
    prev_submission.assignment = submission.assignment
    grader = submissions.resolve_grader(prev_submission)
    if grader:
      graders.append(grader)

  return graders
def summarize_assignment_groups(canvas, args):
  """
  Summarize assignment groups into a single grade:
  - canvas is a Canvas object,
  - args is the command-line arguments, as parsed by argparse.
  """

  try:
    summary = importlib.import_module(args.summary_module)
  except ModuleNotFoundError:
    module_path = pathlib.Path.cwd() / args.summary_module
    module = module_path.stem

    try:
      loader = importlib.machinery.SourceFileLoader(
        module, str(module_path))
      spec = importlib.util.spec_from_loader(module, loader)
      summary = importlib.util.module_from_spec(spec)
      loader.exec_module(summary)
    except Exception as err:
      canvaslms.cli.err(1, f"Error loading summary module "
        f"'{args.summary_module}': {err}")

  courses_list = courses.process_course_option(canvas, args)
  all_assignments = list(assignments.process_assignment_option(canvas, args))

  for course in courses_list:
    users_list = list(course.get_users(enrollment_type=["student"]))
    ag_list = assignments.filter_assignment_groups(
      course, args.assignment_group)

    for assignment_group in ag_list:
      assignments_list = list(assignments.filter_assignments_by_group(
        assignment_group, all_assignments))
      for user, grade, grade_date, *graders in summary.summarize_group(
        assignments_list, users_list):
          yield [
            course.course_code,
            assignment_group.name,
            user.integration_id,
            grade,
            grade_date,
            *graders
          ]

def add_command(subp):
  """Adds the results command to argparse parser subp"""
  results_parser = subp.add_parser("results",
      help="Lists results of a course",
      description="""Lists results of a course for export, for instance to the `ladok report` 
                     command. Output format, CSV:

                       <course code> <component code> <student ID> <grade> <grade date> <graders ...>""",
      epilog="""If you specify an assignment group, the results of the assignments in that 
                group will be summarized. You can supply your own function for summarizing 
                grades through the -S option. See `pydoc3 canvaslms.grades` for different 
                options.""")
  results_parser.set_defaults(func=results_command)
  assignments.add_assignment_option(results_parser, ungraded=False)
  default_summary_module = "canvaslms.grades.conjunctavg"
  results_parser.add_argument("-S", "--summary-module",
    required=False, default=default_summary_module,
    help=f"""Name of Python module or file containing module to load with a custom 
             summarization function to summarize assignment groups. The default module is 
             part of the `canvaslms` package: `{default_summary_module}`. This module must 
             contain a function `summarize_group(assignments, users)`, where `assignments` 
             is a list of `canvasapi.assignment.Assignment` objects and `users` is a list of 
             `canvasapi.user.User` objects. The return value must be a list of lists of the 
             form `[user object, grade, grade date, grader 1, grader 2, ...]`.""")
  results_parser.add_argument("-F", "--include-Fs",
    required=False, default=False, action="store_true",
    help="Include failing grades (Fs) in output. By default we only output "
      "A--Es and Ps.")
